// Netify Agent Device Discovery Processor
// Copyright (C) 2024 eGloo Incorporated
// <http://www.egloo.ca>

#pragma once

#include <condition_variable>
#include <deque>
#include <map>
#include <mutex>
#include <queue>
#include <set>
#include <string>

#include <nd-napi.hpp>
#include <nd-plugin.hpp>

#define _NPP_MAX_CONFIDENCE    80
#define _NPP_MAX_HTTP_UA       10
#define _NPP_MAX_SSDP_UA       10
#define _NPP_MAX_MDNS_SERVICES 10

class nppGlobalConfig
{
public:
    size_t max_confidence;
    size_t max_http_user_agents;
    size_t max_ssdp_user_agents;
    size_t max_mdns_services;
    bool process_all_macs;

    bool api_enabled;
    std::string api_url, api_key;

    nppGlobalConfig() { Reset(); }

    void Reset(void) {
        max_confidence = _NPP_MAX_CONFIDENCE;
        max_http_user_agents = _NPP_MAX_HTTP_UA;
        max_ssdp_user_agents = _NPP_MAX_SSDP_UA;
        max_mdns_services = _NPP_MAX_MDNS_SERVICES;

        process_all_macs = false;

        api_enabled = false;

        api_url.clear();
        api_key.clear();
    }

    void Load(const nlohmann::json &j);
};

class nppChannelConfig
{
public:
    enum Format {
        FORMAT_RAW,
        FORMAT_JSON,
        FORMAT_MSGPACK,
    };

    enum Compressor {
        COMPRESSOR_NONE,
        COMPRESSOR_GZ,
    };

    void Load(const std::string &channel,
      const nlohmann::json &jconf);
    inline void Load(const std::string &channel,
      const nlohmann::json &jconf, nppChannelConfig &defaults) {
        format = defaults.format;
        compressor = defaults.compressor;
        Load(channel, jconf);
    }

    std::string channel;
    Format format = { FORMAT_JSON };
    Compressor compressor = { COMPRESSOR_NONE };
    bool flush = { false };
};

class nppDevice
{
public:
    nppDevice() { }
    nppDevice(const std::string &mac, bool cached = false)
      : mac(mac), cached(cached) { }

    void Update(const nppGlobalConfig &config,
      const nppDevice &device);

    enum EncodeFlags {
        ET_DHCP = 0x01,
        ET_MDNS = 0x02,
        ET_HTTP = 0x04,
        ET_SSDP = 0x08,
        ET_CONFIDENCE = 0x10,
        ET_IP_COUNT = 0x20,
        ET_RESPONSE = 0x40,
        ET_METADATA = (ET_DHCP | ET_MDNS | ET_HTTP | ET_SSDP),
    };

    void Encode(nlohmann::json &j, uint8_t flags = ET_METADATA);

    std::string mac;
    std::string label;
    std::string dhcp_fingerprint;
    std::string dhcp_vendor;
    std::string dhcp_hostname;

    typedef std::deque<std::string> UniqueSet;

    UniqueSet mdns_services;
    UniqueSet http_user_agents;
    UniqueSet ssdp_user_agents;

    typedef std::unordered_set<std::string> IPList;
    IPList ip_list;

    unsigned updates = { 1 };
    unsigned confidence = { 0 };

    bool cached = { false };
    nlohmann::json jresponse;

    bool other_device = { false };
};

class nppPlugin;

class nppApiDeviceDiscovery : public ndNetifyApiThread
{
public:
    nppApiDeviceDiscovery(const nppGlobalConfig &config)
      : config(config), method(ndNetifyApiThread::Method::GET) { }

    nppApiDeviceDiscovery(const nppGlobalConfig &config,
      const nlohmann::json &json)
      : config(config), method(ndNetifyApiThread::Method::POST) {

        payload = json.dump(ndGC_DEBUG ? ND_JSON_INDENT : -1,
          ' ', true, nlohmann::json::error_handler_t::replace);
        mime_type = "application/json";
    };

    virtual void *Entry(void);

protected:
    friend class nppPlugin;

    const nppGlobalConfig &config;
    ndNetifyApiThread::Method method;
    std::string payload;
    std::string mime_type;
};

class nppPlugin : public ndPluginProcessor
{
public:
    nppPlugin(const std::string &tag, const ndPlugin::Params &params);
    virtual ~nppPlugin();

    virtual void *Entry(void);

    virtual void GetVersion(std::string &version);

    template <class T>
    void GetStatus(T &output) const {
        nd_dprintf("%s: TODO\n", __PRETTY_FUNCTION__);
        ndPluginProcessor::GetStatus(output);
    }

    virtual void DispatchEvent(ndPlugin::Event event,
      void *param = nullptr);

    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, ndFlowMap *flow_map);
    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, nd_flow_ptr &flow);
    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, ndInterfaces *interfaces);
    virtual void DispatchProcessorEvent(ndPluginProcessor::Event event,
      const std::string &iface, ndPacketStats *stats);
    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, ndPacketStats *stats);
    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, ndInstanceStatus *status);
    virtual void
    DispatchProcessorEvent(ndPluginProcessor::Event event);

protected:
    std::atomic<bool> reload;
    std::atomic<bool> update;

    void Reload(void);

    nppGlobalConfig globals;
    nppChannelConfig defaults;

    typedef std::map<std::string, std::map<std::string, nppChannelConfig>> Sinks;
    Sinks sinks;

    typedef std::map<std::string, nppDevice> Devices;
    Devices devices;

    typedef std::queue<nd_flow_ptr> FlowQueue;
    FlowQueue flow_queue;

    std::mutex flow_queue_mutex;
    std::condition_variable flow_queue_cond;

    typedef std::set<std::string> Identified;
    Identified identified;

    void ProcessFlow(const nd_flow_ptr &flow);
    void PrintDeviceDebug(const nppDevice &device) const;
    void ProcessUpdate(void);

    void DispatchPayload(nlohmann::json &j);

    nppApiDeviceDiscovery *api_thread;

    void ProcessApiResponse(void);
};
